// https://github.com/meekrosoft/fff
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include <pthread.h>

// Want to keep the argument history for 512 calls
#define FFF_ARG_HISTORY_LEN 512
// Want to keep the call sequence history for 512 function calls
#define FFF_CALL_HISTORY_LEN 512

#include "fff/gtest/gtest.h"
#include "fff/fff.h"
DEFINE_FFF_GLOBALS;

void *gp_blk = (void *)NULL; // 送信するメモリプールのアドレスの確認用

long loop_num; // 無限ループ実行回数
jmp_buf buf;   // 無限ループ脱出用のスタック退避領域

extern pthread_mutex_t mutex; // テストケースを逐次実行するための保険

extern "C"
{
// テスト対象のCソースファイルを丸ごとインクルードすることにより、
// static関数やstatic変数の参照・更新が可能となる。
#include "../base_time_cyclic.c"
}

/* 関数の偽装宣言 */
/* 関数の実体があるものはテスト対象のCソースファイルに記載の関数のみです。 */
/* 実体のない関数は、FFFで偽装します。 */
/* サービスコールの偽装宣言 */
/* タスク付属同期機能 */
FAKE_VALUE_FUNC(ER, slp_tsk);
FAKE_VALUE_FUNC(ER, iwup_tsk, ID);
/* 同期・通信機能 */
FAKE_VALUE_FUNC(ER, psnd_dtq, ID, intptr_t);
/* メモリプール管理機能 */
FAKE_VALUE_FUNC(ER, get_mpf, ID, void **);

/* ライブラリの偽装宣言 */
FAKE_VOID_FUNC(t_perror, uint_t, const char *, int_t, const char *, ER);

// カスタムFake関数の実装
ER get_mpf_custom_fake(ID mpfid, void **p_blk)
{
	*p_blk = malloc(sizeof(64)); // コンフィギュレータで指定する固定長メモリプールのサイズ

	gp_blk = *p_blk; // 送信するアドレスの検証用に保持

	return (E_OK);
}

ER psnd_dtq_custom_fake(ID dtqid, intptr_t data)
{
	EXPECT_EQ(gp_blk, (void *)data);						 // 送信するアドレスの検証
	EXPECT_EQ(EV_TIMER_BASE_TIME, ((T_EVENT *)data)->event); // メッセージのIDを検証

	free((void *)data);
	printf("偽装psnd_dtq経由で無限ループ脱出!\n");
	longjmp(buf, 1);

	printf("無限ループ脱出したのでここは実行されない\n");

	return (E_OK);
}

/* List of fakes used by this unit tester */
#define FFF_FAKES_LIST(FAKE) \
	FAKE(slp_tsk)            \
	FAKE(iwup_tsk)           \
	FAKE(get_mpf)            \
	FAKE(psnd_dtq)

/* テストケースの初期化と後処理 									*/
/* テスト対象、及びテストケースはグローバル変数を使用している。 	*/
/* マルチスレッド環境下で複数のテストケースが同時実行された場合、	*/
/* グローバル変数のアクセスでテストが破綻する。  					*/
/* 複数のテストケースを同時実行することは無いとは思うが、 			*/
/* 保険としてmutexにてテストケースを逐次実行するようにした。		*/
class testBaseTimeCyclic : public testing::Test
{
	void SetUp()
	{
		pthread_mutex_lock(&mutex);
		/* Register resets */
		FFF_FAKES_LIST(RESET_FAKE);
		/* reset common FFF internal structures */
		FFF_RESET_HISTORY();
	}
	void TearDown()
	{
		pthread_mutex_unlock(&mutex);
	}
};

// ベースタイムハンドラのテスト
TEST_F(testBaseTimeCyclic, test_base_time_handler)
{
	// テスト対象をコール
	base_time_handler(0);

	// iwup_tskがコールされていることを確認
	EXPECT_EQ(1, iwup_tsk_fake.call_count);

	// iwup_tskの第1引数が正しいことを確認
	EXPECT_EQ(BASE_TIME_TASK, iwup_tsk_fake.arg0_val);
}

// ベースタイムハンドラから起床されるベースタイムタスクのテスト
TEST_F(testBaseTimeCyclic, test_base_time_task)
{
	// get_mpfをカスタムフェイク関数に設定
	get_mpf_fake.custom_fake = get_mpf_custom_fake;

	// psnd_dtqをカスタムフェイク関数に設定
	psnd_dtq_fake.custom_fake = psnd_dtq_custom_fake;

	// blink点滅周期タスクのメインルーチン(無限ループ)へ突入
	if (0 == setjmp(buf))
	{
		base_time_task(0);
	}
	else
	{
		printf("無限ループ脱出成功 \n\n");
	}

	// slp_tskがコールされていることを確認
	EXPECT_EQ(1, slp_tsk_fake.call_count);

	// get_mpfがコールされていることを確認
	EXPECT_EQ(1, get_mpf_fake.call_count);

	// get_mpfの第1引数が正しいことを確認
	EXPECT_EQ(EVENT_MPF, get_mpf_fake.arg0_val);

	// 第2引数はローカル変数へのポインタのため、ここでは確認できない。
	// psnd_dtq_custom_fake内でメッセージのメンバを確認する
}
