// https://github.com/meekrosoft/fff
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include <pthread.h>

// Want to keep the argument history for 512 calls
#define FFF_ARG_HISTORY_LEN 512
// Want to keep the call sequence history for 512 function calls
#define FFF_CALL_HISTORY_LEN 512

#include "fff/gtest/gtest.h"
#include "fff/fff.h"
DEFINE_FFF_GLOBALS
;

#define TEST_SCENARIO_MAX 100  		// テストシナリオ最大数

const uint32_t *pScenario;				// シナリオ定義を参照するためのポインタ
intptr_t gp_data = (intptr_t) NULL;		// 受信したメモリプールの解放アドレスの確認用

long loop_num;	 						// 無限ループ実行回数
jmp_buf buf; 							// 無限ループ脱出用のスタック退避領域

extern pthread_mutex_t mutex;			// テストケースを逐次実行するための保険

extern "C" {
// テスト対象のCソースファイルを丸ごとインクルードすることにより、
// static関数やstatic変数の参照・更新が可能となる。
#include "../timer_main.c"
}

/* 関数の偽装宣言 */
/* 関数の実体があるものはテスト対象のCソースファイルに記載の関数のみです。 */
/* 実体のない関数は、FFFで偽装します。 */

/* 同期・通信機能 */
FAKE_VALUE_FUNC(ER, rcv_dtq, ID, intptr_t*);
/* メモリプール管理機能 */
FAKE_VALUE_FUNC(ER, rel_mpf, ID, void*);

/* ライブラリの偽装宣言 */
FAKE_VOID_FUNC(t_perror, uint_t, const char*, int_t, const char*, ER);

/* アクション関数の偽装宣言 */
FAKE_VOID_FUNC(StOffEvSw1Off);
FAKE_VOID_FUNC(StOnEvSw1Off);
FAKE_VOID_FUNC(StOffEvSw1On);
FAKE_VOID_FUNC(StOnEvPush1On);
FAKE_VOID_FUNC(StOnEvBaseTime);
FAKE_VOID_FUNC(StOffEntry);
FAKE_VOID_FUNC(StOnEntry);
FAKE_VOID_FUNC(StOffExit);
FAKE_VOID_FUNC(StOnExit);

// カスタムFake関数の実装
ER rel_mpf_custom_fake(ID mpfid, void *blk) {
	EXPECT_EQ(gp_data, (intptr_t )blk);				// 解放するアドレスの検証
	free(blk);
	return (E_OK);
}

ER rcv_dtq_custom_fake(ID dtqid, intptr_t *p_data) {
	if (loop_num < rcv_dtq_fake.call_count) {
		printf("偽装rcv_dtq経由で無限ループ脱出!\n");
		longjmp(buf, 1);

		printf("無限ループ脱出したのでここは実行されない\n");
	}
	/* データキュー受信の偽装 */
	*p_data = (intptr_t) malloc(sizeof(T_EVENT));
	gp_data = *p_data;								// 解放するアドレスの検証用に保持

	((T_EVENT*) *p_data)->event = *(pScenario + (rcv_dtq_fake.call_count) - 1);

	return (E_OK);
}

void StOffEvSw1Off_custom_fake(void)
{
	GTM_NEXTSTS = ST_TIMER_ON; /* 次状態設定			*/
	GTM_CTLF |= DTM_STSCHG;			/* 状態遷移有りを設定	*/
}

/* List of fakes used by this unit tester */
#define FFF_FAKES_LIST(FAKE)	\
        FAKE(rcv_dtq)			\
        FAKE(rel_mpf)			\
		FAKE(t_perror)			\
		FAKE(StOffEvSw1Off)		\
		FAKE(StOnEvSw1Off)		\
		FAKE(StOffEvSw1On)		\
		FAKE(StOnEvPush1On)		\
		FAKE(StOnEvBaseTime)	\
		FAKE(StOffEntry)		\
		FAKE(StOnEntry)			\
		FAKE(StOffExit)			\
		FAKE(StOnExit)

/* テストケースの初期化と後処理 									*/
/* テスト対象、及びテストケースはグローバル変数を使用している。 	*/
/* マルチスレッド環境下で複数のテストケースが同時実行された場合、	*/
/* グローバル変数のアクセスでテストが破綻する。  					*/
/* 複数のテストケースを同時実行することは無いとは思うが、 			*/
/* 保険としてmutexにてテストケースを逐次実行するようにした。		*/
class testTimerMain: public testing::Test {
	void SetUp() {
		pthread_mutex_lock(&mutex);
		/* Register resets */
		FFF_FAKES_LIST(RESET_FAKE);
		/* reset common FFF internal structures */
		FFF_RESET_HISTORY();
	}
	void TearDown() {
		pthread_mutex_unlock(&mutex);
	}
};

// 静的API（ATT_INI）に登録した初期化ルーチンのテスト
TEST_F(testTimerMain, test_timer_task_init) {
	memset(&gtm_ctl, 0xFF, sizeof(TTM_CTL));	// 予めメモリを汚しておく

	timer_task_init(0);						// テスト対象をコール

	// 初期化後の期待値を確認
	EXPECT_EQ(0, gtm_ctl.ctl_flg);
	EXPECT_EQ(0, gtm_ctl.timer);
	EXPECT_EQ(ST_TIMER_OFF, gtm_ctl.nowsts);
	EXPECT_EQ(ST_TIMER_OFF, gtm_ctl.nextsts);
}

// イベント番号が範囲内かをチェックする関数のテスト
TEST_F(testTimerMain, testCheckTimerEvent) {
	T_EVENT test_msg;

	gtm_ctl.prcv_data = &test_msg;

	test_msg.event = EV_TIMER_SW1_OFF;
	EXPECT_TRUE(check_timer_event());

	test_msg.event = EV_TIMER_SW1_ON;
	EXPECT_TRUE(check_timer_event());

	test_msg.event = EV_TIMER_PUSH1_ON;
	EXPECT_TRUE(check_timer_event());

	test_msg.event = EV_TIMER_BASE_TIME;
	EXPECT_TRUE(check_timer_event());

	test_msg.event = EV_TIMER_NUM;
	EXPECT_FALSE(check_timer_event());
}

// Timerタスクのメインルーチンのテストケースのテンプレート
TEST_F(testTimerMain, testTEMPLATE) {
	int i;
	// テストケースのテンプレート
	// 変更が必要な箇所のみ「要変更」とコメントしました
	// 他は基本的に変更しません

	// シナリオの作成
	static const uint32_t scenario[] = { EV_TIMER_SW1_OFF };
	// ↑要変更 テスト用のシナリオです。
	// この配列にEV_BLINKで定義されるメッセージを追加していきます。
	// 配列順にメッセージが送信されます。
	// 配列の最大要素数はTEST_SCENARIO_MAXで調整します。

	// 静的API（ATT_INI）により登録した初期化ルーチンは自動的に実行されないため、ここで実行する
	timer_task_init(0);

	loop_num = sizeof scenario / sizeof scenario[0];	// 無限ループ実行回数
	rcv_dtq_fake.custom_fake = rcv_dtq_custom_fake;		// rcv_dtqをカスタムフェイク関数に設定
	rel_mpf_fake.custom_fake = rel_mpf_custom_fake; 	// rel_mpfをカスタムフェイク関数に設定

	pScenario = &scenario[0];							// シナリオ配列へのポインタ取得

	if (0 == setjmp(buf)) {
		timer_task(0);							// blinkタスクのメインルーチン(無限ループ)へ突入
	} else {
		printf("無限ループ脱出成功 \n\n");
	}

	i = 0;	// i++より即値のほうが効率的だが、テストケースなので良しとする
	// ↓要変更 シナリオが期待する判定文を記述してください。
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* )StOffEvSw1Off);
	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// 最後、偽装したrcv_dtq内のlongjmpにより、タスクのメインルーチンを脱出
	EXPECT_EQ(fff.call_history[i], (void* )rcv_dtq);
}

// Blinkタスクのメインルーチンの状態遷移のテスト
TEST_F(testTimerMain, testSTSCHG) {
	int i;

	// シナリオの作成
	static const uint32_t scenario[] = { EV_TIMER_SW1_OFF };

	// 静的API（ATT_INI）により登録した初期化ルーチンは自動的に実行されないため、ここで実行する
	timer_task_init(0);

	loop_num = sizeof scenario / sizeof scenario[0];	// 無限ループ実行回数
	rcv_dtq_fake.custom_fake = rcv_dtq_custom_fake;		// rcv_dtqをカスタムフェイク関数に設定
	rel_mpf_fake.custom_fake = rel_mpf_custom_fake; 	// rel_mpfをカスタムフェイク関数に設定

	// カスタムフェイク関数内で状態遷移有りを設定する
	StOffEvSw1Off_fake.custom_fake = StOffEvSw1Off_custom_fake;

	pScenario = &scenario[0];							// シナリオ配列へのポインタ取得

	if (0 == setjmp(buf)) {
		timer_task(0);							// blinkタスクのメインルーチン(無限ループ)へ突入
	} else {
		printf("無限ループ脱出成功 \n\n");
	}

	i = 0;
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* )StOffEvSw1Off);

	EXPECT_EQ(ST_TIMER_ON, GTM_NOWSTS);				// 状態遷移したことを確認

	EXPECT_EQ(fff.call_history[i++], (void* )StOffExit);
	EXPECT_EQ(fff.call_history[i++], (void* )StOnEntry);

	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// 最後、偽装したrcv_dtq内のlongjmpにより、タスクのメインルーチンを脱出
	EXPECT_EQ(fff.call_history[i], (void* )rcv_dtq);
}

// Timerタスクの初期状態(ST_TIMER_OFF)での全イベント受信のテストケース
TEST_F(testTimerMain, test_State_TIMER_OFF) {
	int i;
	// シナリオの定義
	static const uint32_t scenario[] = {
		EV_TIMER_SW1_OFF,	/* スライドスイッチOFF通知 						*/
		EV_TIMER_SW1_ON, 	/* スライドスイッチONライドスイッチON通知		*/
		EV_TIMER_PUSH1_ON, 	/* PUSH1ON通知 									*/
		EV_TIMER_BASE_TIME, /* 1秒経過 										*/
		EV_TIMER_NUM 		/* イベント番号が不正(上限超過)					*/
	};

	timer_task_init(0);

	loop_num = sizeof scenario / sizeof scenario[0];	// 無限ループ実行回数
	rcv_dtq_fake.custom_fake = rcv_dtq_custom_fake;		// rcv_dtqをカスタムフェイク関数に設定
	rel_mpf_fake.custom_fake = rel_mpf_custom_fake; 	// rel_mpfをカスタムフェイク関数に設定

	pScenario = &scenario[0];							// シナリオ配列へのポインタ取得

	if (0 == setjmp(buf)) {
		timer_task(0);							// blinkタスクのメインルーチン(無限ループ)へ突入
	} else {
		printf("無限ループ脱出成功 \n\n");
	}

//	EXPECT_EQ(1, stp_cyc_fake.call_count);

	i = 0;
	// スライドスイッチOFF通知
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* )StOffEvSw1Off);
	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// スライドスイッチON通知
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* )StOffEvSw1On);
	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// PUSH1ON通知
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// 1秒経過
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// イベント番号が不正(上限超過)
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// 最後、偽装したrcv_dtq内のlongjmpにより、タスクのメインルーチンを脱出
	EXPECT_EQ(fff.call_history[i], (void* )rcv_dtq);
}

// タイマタスクのタイマON状態(ST_TIMER_ONでの全イベント受信のテストケース
TEST_F(testTimerMain, test_STSTE_TIMER_ON) {
	int i;
	// シナリオの定義
	static const uint32_t scenario[] = {
			EV_TIMER_SW1_OFF, 	/* スライドスイッチOFF通知 					*/
			EV_TIMER_SW1_ON, 	/* スライドスイッチONライドスイッチON通知	*/
			EV_TIMER_PUSH1_ON, 	/* PUSH1ON通知 								*/
			EV_TIMER_BASE_TIME, /* 1秒経過 									*/
			EV_TIMER_NUM 		/* イベント番号が不正(上限超過)				*/
	};

	timer_task_init(0);
	gtm_ctl.nowsts = ST_TIMER_ON;

	loop_num = sizeof scenario / sizeof scenario[0];	// 無限ループ実行回数
	rcv_dtq_fake.custom_fake = rcv_dtq_custom_fake;		// rcv_dtqをカスタムフェイク関数に設定
	rel_mpf_fake.custom_fake = rel_mpf_custom_fake; 	// rel_mpfをカスタムフェイク関数に設定

	pScenario = &scenario[0];							// シナリオ配列へのポインタ取得

	if (0 == setjmp(buf)) {
		timer_task(0);							// blinkタスクのメインルーチン(無限ループ)へ突入
	} else {
		printf("無限ループ脱出成功 \n\n");
	}

	i = 0;
	// スライドスイッチOFF通知
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* )StOnEvSw1Off);
	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// スライドスイッチON通知
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// PUSH1ON通知
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* ) StOnEvPush1On);
	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// 1秒経過
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* )StOnEvBaseTime);
	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// イベント番号が不正(上限超過)
	EXPECT_EQ(fff.call_history[i++], (void* )rcv_dtq);
	EXPECT_EQ(fff.call_history[i++], (void* )rel_mpf);

	// 最後、偽装したrcv_dtq内のlongjmpにより、タスクのメインルーチンを脱出
	EXPECT_EQ(fff.call_history[i], (void* )rcv_dtq);
}

